# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# column.py
#
# Implements the Column class as documented in DataFrame API
# -----------------------------------------------------------------------------

from typing import Any, Dict, Iterable, Optional, Tuple

from .buffer import OracleColumnBuffer
from .protocol import (
    CategoricalDescription,
    Column,
    Dtype,
    ColumnBuffers,
    ColumnNullType,
    DtypeKind,
)

from .nanoarrow_bridge import (
    ArrowTimeUnit,
    ArrowType,
)


class OracleColumn(Column):
    """
    OracleColumn represents a column in the DataFrame Interchange Protocol. It
    provides a standardized way to expose a column's data, metadata and chunks,
    allowing interoperability between data frame libraries.
    """

    def __init__(self, ora_arrow_array: object):
        self.ora_arrow_array = ora_arrow_array
        self._buffer_info = ora_arrow_array.get_buffer_info()

    def __arrow_c_array__(self, requested_schema=None):
        return self.ora_arrow_array.__arrow_c_array__(
            requested_schema=requested_schema
        )

    def _data_buffer(self):
        buffer = self._buffer_info.get("data")
        if buffer is None:
            return None
        size_bytes, address = buffer
        data_buffer = OracleColumnBuffer(
            size_in_bytes=size_bytes, address=address, buffer_type="data"
        )
        return data_buffer, self.dtype

    def _offsets_buffer(self):
        buffer = self._buffer_info.get("offsets")
        if buffer is None:
            return None
        size_bytes, address = buffer
        offsets_buffer = OracleColumnBuffer(
            size_in_bytes=size_bytes, address=address, buffer_type="offsets"
        )
        if self.ora_arrow_array.arrow_type in (
            ArrowType.NANOARROW_TYPE_LARGE_STRING,
            ArrowType.NANOARROW_TYPE_LARGE_BINARY,
        ):
            dtype = (DtypeKind.INT, 64, "l", "=")
        else:
            dtype = (DtypeKind.INT, 32, "i", "=")
        return offsets_buffer, dtype

    def _validity_buffer(self):
        buffer = self._buffer_info.get("validity")
        if buffer is None:
            return None
        size_bytes, address = buffer
        validity_buffer = OracleColumnBuffer(
            size_in_bytes=size_bytes, address=address, buffer_type="validity"
        )
        dtype = (DtypeKind.BOOL, 1, "b", "=")
        return validity_buffer, dtype

    def describe_categorical(self) -> CategoricalDescription:
        """
        Returns a description of a categorical data type.
        """
        raise NotImplementedError()

    @property
    def describe_null(self) -> Tuple[ColumnNullType, Optional[int]]:
        """
        Returns a description of the null representation used by the column.
        """
        if self.null_count == 0:
            return ColumnNullType.NON_NULLABLE, None
        else:
            return ColumnNullType.USE_BITMASK, 0

    @property
    def dtype(self) -> Dtype:
        """
        Returns the data type of the column. The returned dtype provides
        information on the storage format and the type of data in the column.
        """
        arrow_type = self.ora_arrow_array.arrow_type
        if arrow_type == ArrowType.NANOARROW_TYPE_INT64:
            return (DtypeKind.INT, 64, "l", "=")
        elif arrow_type == ArrowType.NANOARROW_TYPE_DOUBLE:
            return (DtypeKind.FLOAT, 64, "g", "=")
        elif arrow_type == ArrowType.NANOARROW_TYPE_FLOAT:
            return (DtypeKind.FLOAT, 64, "g", "=")
        elif arrow_type == ArrowType.NANOARROW_TYPE_STRING:
            return (DtypeKind.STRING, 8, "u", "=")
        elif arrow_type == ArrowType.NANOARROW_TYPE_TIMESTAMP:
            time_unit = self.ora_arrow_array.time_unit
            if time_unit == ArrowTimeUnit.NANOARROW_TIME_UNIT_MICRO:
                return (DtypeKind.DATETIME, 64, "tsu:", "=")
            elif time_unit == ArrowTimeUnit.NANOARROW_TIME_UNIT_SECOND:
                return (DtypeKind.DATETIME, 64, "tss:", "=")
            elif time_unit == ArrowTimeUnit.NANOARROW_TIME_UNIT_MILLI:
                return (DtypeKind.DATETIME, 64, "tsm:", "=")
            elif time_unit == ArrowTimeUnit.NANOARROW_TIME_UNIT_NANO:
                return (DtypeKind.DATETIME, 64, "tsn:", "=")
        elif arrow_type == ArrowType.NANOARROW_TYPE_DECIMAL128:
            array = self.ora_arrow_array
            return (
                DtypeKind.DECIMAL,
                128,
                f"d:{array.precision}.{array.scale}",
                "=",
            )
        elif arrow_type == ArrowType.NANOARROW_TYPE_BINARY:
            return (DtypeKind.STRING, 8, "z", "=")
        elif arrow_type == ArrowType.NANOARROW_TYPE_LARGE_BINARY:
            return (DtypeKind.STRING, 8, "Z", "=")
        elif arrow_type == ArrowType.NANOARROW_TYPE_LARGE_STRING:
            return (DtypeKind.STRING, 8, "U", "=")

    def get_buffers(self) -> ColumnBuffers:
        """
        Returns a dictionary specifying the memory buffers backing the column.
        This currently consists of:
        - "data": the main buffer storing column values
        - "validity": a buffer containing null/missing values
        - "offsets": a buffer for variable-length types like string
        """
        return {
            "data": self._data_buffer(),
            "validity": self._validity_buffer(),
            "offsets": self._offsets_buffer(),
        }

    def get_chunks(self, n_chunks: Optional[int] = None) -> Iterable[Column]:
        """
        Return an iterator containing the column chunks. Currently this only
        returns itself.
        """
        yield self

    @property
    def metadata(self) -> Dict[str, Any]:
        """
        Returns metadata about the column.
        """
        return {
            "name": self.ora_arrow_array.name,
            "size": self.size(),
            "num_chunks": self.num_chunks(),
        }

    @property
    def null_count(self) -> int:
        """
        Returns the number of null elements.
        """
        return self.ora_arrow_array.null_count

    def num_chunks(self) -> int:
        """
        Returns the number of chunks used by the column. This method currently
        always returns the value 1, implying that the column uses contiguous
        memory.
        """
        return 1

    @property
    def offset(self) -> int:
        """
        Returns the offset of the first element.
        """
        return self.ora_arrow_array.offset

    def size(self) -> int:
        """
        Returns the number of elements in the column.
        """
        return len(self.ora_arrow_array)



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
1000 - Module for static type checking of the cursor module.
"""

import oracledb


def stc_1000(cursor: oracledb.Cursor):
    cursor.execute(None, parameters=None)
    cursor.execute("sql", parameters=None)
    cursor.execute("sql", parameters=[1, 2, 3])
    cursor.execute("sql", parameters=(1, 2, 3))
    cursor.execute("sql", parameters=dict(arg1=5, arg2=6))


def stc_1001(cursor: oracledb.Cursor):
    cursor.callproc("name", parameters=None)
    cursor.callproc("name", parameters=(1, 2, 3))
    cursor.callproc("name", parameters=[1, 2, 3])
    cursor.callproc("name", keyword_parameters=None)
    cursor.callproc("name", keyword_parameters=dict(arg1=5, arg2=6))
    cursor.callproc("name", keywordParameters=None)
    cursor.callproc("name", keywordParameters=dict(arg1=5, arg2=6))


def stc_1002(cursor: oracledb.Cursor):
    cursor.callfunc("name", str, parameters=None)
    cursor.callfunc("name", str, parameters=(1, 2, 3))
    cursor.callfunc("name", str, parameters=[1, 2, 3])
    cursor.callfunc("name", str, keyword_parameters=None)
    cursor.callfunc("name", str, keyword_parameters=dict(arg1=5, arg2=6))
    cursor.callfunc("name", str, keywordParameters=None)
    cursor.callfunc("name", str, keywordParameters=dict(arg1=5, arg2=6))


def stc_1003(cursor: oracledb.Cursor):
    cursor.executemany(None, [(1, 2, 3), (4, 5, 6)])
    cursor.executemany("sql", [(1, 2, 3), (4, 5, 6)])
    cursor.executemany("sql", 8)


def stc_1004(cursor: oracledb.Cursor):
    cursor.fetchmany(size=None)
    cursor.fetchmany(size=5)
    cursor.fetchmany(numRows=None)
    cursor.fetchmany(numRows=5)


async def stc_1005(cursor: oracledb.AsyncCursor):
    await cursor.callfunc("name", str, parameters=None)
    await cursor.callfunc("name", str, parameters=(1, 2, 3))
    await cursor.callfunc("name", str, parameters=[1, 2, 3])
    await cursor.callfunc("name", str, keyword_parameters=None)
    await cursor.callfunc("name", str, keyword_parameters=dict(arg1=5, arg2=6))


async def stc_1006(cursor: oracledb.AsyncCursor):
    await cursor.callproc("name", parameters=None)
    await cursor.callproc("name", parameters=(1, 2, 3))
    await cursor.callproc("name", parameters=[1, 2, 3])
    await cursor.callproc("name", keyword_parameters=None)
    await cursor.callproc("name", keyword_parameters=dict(arg1=5, arg2=6))


async def stc_1007(cursor: oracledb.AsyncCursor):
    await cursor.execute(None, parameters=None)
    await cursor.execute("sql", parameters=None)
    await cursor.execute("sql", parameters=[1, 2, 3])
    await cursor.execute("sql", parameters=(1, 2, 3))
    await cursor.execute("sql", parameters=dict(arg1=5, arg2=6))


async def stc_1008(cursor: oracledb.AsyncCursor):
    await cursor.executemany(None, [(1, 2, 3), (4, 5, 6)])
    await cursor.executemany("sql", [(1, 2, 3), (4, 5, 6)])
    await cursor.executemany("sql", 8)


async def stc_1009(cursor: oracledb.AsyncCursor):
    await cursor.fetchmany(size=None)
    await cursor.fetchmany(size=5)



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
8200 - Module for testing AQ Bulk enqueue/dequeue with asyncio
"""

import datetime
import threading
import unittest

import oracledb
import test_env

RAW_QUEUE_NAME = "TEST_RAW_QUEUE"
JSON_QUEUE_NAME = "TEST_JSON_QUEUE"
RAW_PAYLOAD_DATA = [
    "The first message",
    "The second message",
    "The third message",
    "The fourth message",
    "The fifth message",
    "The sixth message",
    "The seventh message",
    "The eighth message",
    "The ninth message",
    "The tenth message",
    "The eleventh message",
    "The twelfth and final message",
]

JSON_DATA_PAYLOAD = [
    [
        2.75,
        True,
        "Ocean Beach",
        b"Some bytes",
        {"keyA": 1.0, "KeyB": "Melbourne"},
        datetime.datetime(2022, 8, 1, 0, 0),
    ],
    dict(name="John", age=30, city="New York"),
]


@unittest.skipUnless(
    test_env.get_is_thin(), "asyncio not supported in thick mode"
)
class TestCase(test_env.BaseAsyncTestCase):
    async def __deq_in_thread(self, results):
        async with test_env.get_connection_async() as conn:
            queue = conn.queue(RAW_QUEUE_NAME)
            queue.deqoptions.wait = 10
            queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
            while len(results) < len(RAW_PAYLOAD_DATA):
                messages = await queue.deqmany(5)
                if not messages:
                    break
                for message in messages:
                    results.append(message.payload.decode())
            await conn.commit()

    async def test_8200(self):
        "8200 - test bulk enqueue and dequeue"
        queue = await self.get_and_clear_queue(RAW_QUEUE_NAME)
        messages = [
            self.conn.msgproperties(payload=data) for data in RAW_PAYLOAD_DATA
        ]
        await queue.enqmany(messages)
        messages = await queue.deqmany(len(RAW_PAYLOAD_DATA))
        data = [message.payload.decode() for message in messages]
        await self.conn.commit()
        self.assertEqual(data, RAW_PAYLOAD_DATA)

    async def test_8201(self):
        "8201 - test empty bulk dequeue"
        queue = await self.get_and_clear_queue(RAW_QUEUE_NAME)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        messages = await queue.deqmany(5)
        await self.conn.commit()
        self.assertEqual(messages, [])

    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support enq immediate yet"
    )
    async def test_8202(self):
        "8202 - test bulk dequeue with wait"
        queue = await self.get_and_clear_queue(RAW_QUEUE_NAME)
        results = []
        thread = threading.Thread(target=self.__deq_in_thread, args=(results,))
        thread.start()
        messages = [
            self.conn.msgproperties(payload=data) for data in RAW_PAYLOAD_DATA
        ]
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        await queue.enqmany(messages)
        thread.join()
        self.assertEqual(results, RAW_PAYLOAD_DATA)

    async def test_8203(self):
        "8203 - test enqueue and dequeue multiple times"
        queue = await self.get_and_clear_queue(RAW_QUEUE_NAME)
        data_to_enqueue = RAW_PAYLOAD_DATA
        for num in (2, 6, 4):
            messages = [
                self.conn.msgproperties(payload=data)
                for data in data_to_enqueue[:num]
            ]
            data_to_enqueue = data_to_enqueue[num:]
            await queue.enqmany(messages)
        await self.conn.commit()
        all_data = []
        for num in (3, 5, 10):
            messages = await queue.deqmany(num)
            all_data.extend(message.payload.decode() for message in messages)
        await self.conn.commit()
        self.assertEqual(all_data, RAW_PAYLOAD_DATA)

    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support enq immediate yet"
    )
    async def test_8204(self):
        "8204 - test visibility option for enqueue and dequeue"
        queue = await self.get_and_clear_queue(RAW_QUEUE_NAME)

        # first test with ENQ_ON_COMMIT (commit required)
        queue.enqoptions.visibility = oracledb.ENQ_ON_COMMIT
        props1 = self.conn.msgproperties(payload="A first message")
        props2 = self.conn.msgproperties(payload="A second message")
        await queue.enqmany([props1, props2])
        async with test_env.get_connection_async() as other_conn:
            other_queue = other_conn.queue(RAW_QUEUE_NAME)
            other_queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
            other_queue.deqoptions.visibility = oracledb.DEQ_ON_COMMIT
            messages = await other_queue.deqmany(5)
            self.assertEqual(len(messages), 0)
            await self.conn.commit()
            messages = await other_queue.deqmany(5)
            self.assertEqual(len(messages), 2)
            await other_conn.rollback()

            # second test with ENQ_IMMEDIATE (no commit required)
            queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
            other_queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
            queue.enqmany([props1, props2])
            messages = await other_queue.deqmany(5)
            self.assertEqual(len(messages), 4)
            await other_conn.rollback()
            messages = await other_queue.deqmany(5)
            self.assertEqual(len(messages), 0)

    async def test_8205(self):
        "8205 - test error for messages with no payload"
        queue = await self.get_and_clear_queue(RAW_QUEUE_NAME)
        messages = [self.conn.msgproperties() for _ in RAW_PAYLOAD_DATA]
        with self.assertRaisesFullCode("DPY-2000"):
            await queue.enqmany(messages)

    async def test_8206(self):
        "8206 - verify that the msgid property is returned correctly"
        queue = await self.get_and_clear_queue(RAW_QUEUE_NAME)
        messages = [
            self.conn.msgproperties(payload=data) for data in RAW_PAYLOAD_DATA
        ]
        await queue.enqmany(messages)
        await self.cursor.execute("select msgid from raw_queue_tab")
        actual_msgids = set(m for m, in await self.cursor.fetchall())
        msgids = set(message.msgid for message in messages)
        self.assertEqual(msgids, actual_msgids)
        messages = await queue.deqmany(len(RAW_PAYLOAD_DATA))
        msgids = set(message.msgid for message in messages)
        self.assertEqual(msgids, actual_msgids)

    async def test_8207(self):
        "4800 - test enqueuing and dequeuing JSON message"
        queue = await self.get_and_clear_queue(JSON_QUEUE_NAME, "JSON")
        props = [
            self.conn.msgproperties(payload=data) for data in JSON_DATA_PAYLOAD
        ]
        await queue.enqmany(props)
        await self.conn.commit()
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        messages = await queue.deqmany(5)
        actual_data = [message.payload for message in messages]
        self.assertEqual(actual_data, JSON_DATA_PAYLOAD)

    async def test_8208(self):
        "8208 - test enqueuing to a JSON queue without a JSON payload"
        queue = await self.get_and_clear_queue(JSON_QUEUE_NAME, "JSON")
        props = self.conn.msgproperties(payload="string message")
        with self.assertRaisesFullCode("DPY-2062"):
            await queue.enqmany([props, props])

    async def test_8209(self):
        "8209 - test errors for invalid values for enqmany and deqmany"
        queue = await self.get_and_clear_queue(JSON_QUEUE_NAME, "JSON")
        props = self.conn.msgproperties(payload="string message")
        with self.assertRaises(TypeError):
            await queue.enqmany(props)
        with self.assertRaises(TypeError):
            await queue.enqmany(["Not", "msgproperties"])
        with self.assertRaises(TypeError):
            await queue.deqmany("5")


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# queue.pyx
#
# Cython file defining the thin Queue implementation class (embedded in
# thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseThinQueueImpl(BaseQueueImpl):

    cdef:
        BaseThinConnImpl _conn_impl
        bytes payload_toid

    cdef AqArrayMessage _create_array_deq_message(self, uint32_t num_iters):
        """
        Create the message used for dequeuing multiple AQ messages
        """
        cdef:
            AqArrayMessage message
            uint32_t i
        message = self._conn_impl._create_message(AqArrayMessage)
        message.num_iters = num_iters
        message.props_impls = [ThinMsgPropsImpl() for i in range(num_iters)]
        message.queue_impl = self
        message.deq_options_impl = self.deq_options_impl
        message.operation = TNS_AQ_ARRAY_DEQ
        return message

    cdef AqArrayMessage _create_array_enq_message(self, list props_impls):
        """
        Create the message used for enqueuing multiple AQ messages
        """
        cdef AqArrayMessage message
        message = self._conn_impl._create_message(AqArrayMessage)
        message.queue_impl = self
        message.enq_options_impl = self.enq_options_impl
        message.props_impls = props_impls
        message.operation = TNS_AQ_ARRAY_ENQ
        message.num_iters = len(props_impls)
        return message

    cdef AqDeqMessage _create_deq_message(self):
        """
        Create the message for dequeuing a payload.
        """
        cdef:
            ThinMsgPropsImpl props_impl
            AqDeqMessage message
        props_impl = ThinMsgPropsImpl()
        message = self._conn_impl._create_message(AqDeqMessage)
        message.queue_impl = self
        message.deq_options_impl = self.deq_options_impl
        message.props_impl = props_impl
        return message

    cdef AqEnqMessage _create_enq_message(self, ThinMsgPropsImpl props_impl):
        """
        Create the message for enqueuing the provided payload.
        """
        cdef AqEnqMessage message
        message = self._conn_impl._create_message(AqEnqMessage)
        message.queue_impl = self
        message.enq_options_impl = self.enq_options_impl
        message.props_impl = props_impl
        return message

    def initialize(self, BaseThinConnImpl conn_impl, str name,
                   ThinDbObjectTypeImpl payload_type, bint is_json):
        """
        Internal method for initializing the queue.
        """
        self._conn_impl = conn_impl
        self.is_json = is_json
        self.deq_options_impl = ThinDeqOptionsImpl()
        self.enq_options_impl = ThinEnqOptionsImpl()
        self.payload_type = payload_type
        if self.is_json:
            self.payload_toid = bytes([0]*15+[0x47])
        elif self.payload_type is not None:
            self.payload_toid = payload_type.oid
        else:
            self.payload_toid = bytes([0]*15+[0x17])
        self.name = name


cdef class ThinQueueImpl(BaseThinQueueImpl):

    def deq_many(self, uint32_t max_num_messages):
        """
        Internal method for dequeuing multiple messages from a queue.
        """
        cdef:
            Protocol protocol = <Protocol> self._conn_impl._protocol
            AqArrayMessage message
        message = self._create_array_deq_message(max_num_messages)
        protocol._process_single_message(message)
        if message.no_msg_found:
            return []
        return message.props_impls[:message.num_iters]

    def deq_one(self):
        """
        Internal method for dequeuing a single message from a queue.
        """
        cdef:
            Protocol protocol = <Protocol> self._conn_impl._protocol
            AqDeqMessage message
        message = self._create_deq_message()
        protocol._process_single_message(message)
        if not message.no_msg_found:
            return message.props_impl

    def enq_many(self, list props_impls):
        """
        Internal method for enqueuing many messages into a queue.
        """
        cdef :
            Protocol protocol = <Protocol> self._conn_impl._protocol
            AqArrayMessage message
        message = self._create_array_enq_message(props_impls)
        protocol._process_single_message(message)

    def enq_one(self, ThinMsgPropsImpl props_impl):
        """
        Internal method for enqueuing a single message into a queue.
        """
        cdef:
            Protocol protocol = <Protocol> self._conn_impl._protocol
            AqEnqMessage message
        message = self._create_enq_message(props_impl)
        protocol._process_single_message(message)


cdef class AsyncThinQueueImpl(BaseThinQueueImpl):

    async def deq_many(self, uint32_t max_num_messages):
        """
        Internal method for dequeuing multiple messages from a queue.
        """
        cdef:
            BaseAsyncProtocol protocol
            AqArrayMessage message
        protocol = <BaseAsyncProtocol> self._conn_impl._protocol
        message = self._create_array_deq_message(max_num_messages)
        await protocol._process_single_message(message)
        if message.no_msg_found:
            return []
        return message.props_impls[:message.num_iters]

    async def deq_one(self):
        """
        Internal method for dequeuing a single message from a queue.
        """
        cdef:
            BaseAsyncProtocol protocol
            AqDeqMessage message
        protocol = <BaseAsyncProtocol> self._conn_impl._protocol
        message = self._create_deq_message()
        await protocol._process_single_message(message)
        if not message.no_msg_found:
            return message.props_impl

    async def enq_many(self, list props_impls):
        """
        Internal method for enqueuing many messages into a queue.
        """
        cdef :
            BaseAsyncProtocol protocol
            AqArrayMessage message
        protocol = <BaseAsyncProtocol> self._conn_impl._protocol
        message = self._create_array_enq_message(props_impls)
        await protocol._process_single_message(message)

    async def enq_one(self, ThinMsgPropsImpl props_impl):
        """
        Internal method for enqueuing a single message into a queue.
        """
        cdef:
            BaseAsyncProtocol protocol
            AqEnqMessage message
        protocol = <BaseAsyncProtocol> self._conn_impl._protocol
        message = self._create_enq_message(props_impl)
        await protocol._process_single_message(message)


cdef class ThinDeqOptionsImpl(BaseDeqOptionsImpl):
    cdef:
        str condition
        str consumer_name
        str correlation
        uint16_t delivery_mode
        uint32_t mode
        bytes msgid
        uint32_t navigation
        str transformation
        uint32_t visibility
        uint32_t wait

    def __init__(self):
        self.delivery_mode = TNS_AQ_MSG_PERSISTENT
        self.mode = TNS_AQ_DEQ_REMOVE
        self.navigation = TNS_AQ_DEQ_NEXT_MSG
        self.visibility = TNS_AQ_DEQ_ON_COMMIT
        self.wait = TNS_AQ_DEQ_WAIT_FOREVER

    def get_condition(self):
        """
        Internal method for getting the condition.
        """
        return self.condition

    def get_consumer_name(self):
        """
        Internal method for getting the consumer name.
        """
        return self.consumer_name

    def get_correlation(self):
        """
        Internal method for getting the correlation.
        """
        return self.correlation

    def get_message_id(self):
        """
        Internal method for getting the message id.
        """
        return self.msgid

    def get_mode(self):
        """
        Internal method for getting the mode.
        """
        return self.mode

    def get_navigation(self):
        """
        Internal method for getting the navigation.
        """
        return self.navigation

    def get_transformation(self):
        """
        Internal method for getting the transformation.
        """
        return self.transformation

    def get_visibility(self):
        """
        Internal method for getting the visibility.
        """
        return self.visibility

    def get_wait(self):
        """
        Internal method for getting the wait.
        """
        return self.wait

    def set_condition(self, str value):
        """
        Internal method for setting the condition.
        """
        self.condition = value

    def set_consumer_name(self, str value):
        """
        Internal method for setting the consumer name.
        """
        self.consumer_name =  value

    def set_correlation(self, str value):
        """
        Internal method for setting the correlation.
        """
        self.correlation = value

    def set_delivery_mode(self, uint16_t value):
        """
        Internal method for setting the delivery mode.
        """
        self.delivery_mode = value

    def set_mode(self, uint32_t value):
        """
        Internal method for setting the mode.
        """
        self.mode = value

    def set_message_id(self, bytes value):
        """
        Internal method for setting the message id.
        """
        self.msgid = value

    def set_navigation(self, uint32_t value):
        """
        Internal method for setting the navigation.
        """
        self.navigation = value

    def set_transformation(self, str value):
        """
        Internal method for setting the transformation.
        """
        self.transformation = value

    def set_visibility(self, uint32_t value):
        """
        Internal method for setting the visibility.
        """
        self.visibility = value

    def set_wait(self, uint32_t value):
        """
        Internal method for setting the wait.
        """
        self.wait = value


cdef class ThinEnqOptionsImpl(BaseEnqOptionsImpl):
    cdef:
        str transformation
        uint32_t visibility
        uint32_t delivery_mode

    def __init__(self):
        self.visibility = TNS_AQ_ENQ_ON_COMMIT
        self.delivery_mode = TNS_AQ_MSG_PERSISTENT

    def get_transformation(self):
        """
        Internal method for getting the transformation.
        """
        return self.transformation

    def get_visibility(self):
        """
        Internal method for getting the visibility.
        """
        return self.visibility

    def set_delivery_mode(self, uint16_t value):
        """
        Internal method for setting the delivery mode.
        """
        self.delivery_mode = value

    def set_transformation(self, str value):
        """
        Internal method for setting the transformation.
        """
        self.transformation = value

    def set_visibility(self, uint32_t value):
        """
        Internal method for setting the visibility.
        """
        self.visibility = value


cdef class ThinMsgPropsImpl(BaseMsgPropsImpl):

    cdef:
        int32_t delay
        str correlation
        str exceptionq
        int32_t expiration
        int32_t priority
        list recipients
        int32_t num_attempts
        uint32_t delivery_mode
        cydatetime.datetime enq_time
        bytes msgid
        int32_t state
        object payload_obj
        BaseThinConnImpl _conn_impl
        bytes enq_txn_id
        bytes sender_agent_name
        bytes sender_agent_address
        unsigned char sender_agent_protocol
        bytes original_msg_id

    def __init__(self):
        self.delay = TNS_AQ_MSG_NO_DELAY
        self.expiration = TNS_AQ_MSG_NO_EXPIRATION
        self.recipients = []
        self.sender_agent_protocol = 0

    def get_num_attempts(self):
        """
        Internal method for getting the number of attempts made.
        """
        return self.num_attempts

    def get_correlation(self):
        """
        Internal method for getting the correlation.
        """
        return self.correlation

    def get_delay(self):
        """
        Internal method for getting the delay.
        """
        return self.delay

    def get_delivery_mode(self):
        """
        Internal method for getting the delivery mode.
        """
        return self.delivery_mode

    def get_enq_time(self):
        """
        Internal method for getting the enqueue time.
        """
        return self.enq_time

    def get_exception_queue(self):
        """
        Internal method for getting the exception queue.
        """
        return self.exceptionq

    def get_expiration(self):
        """
        Internal method for getting the message expiration.
        """
        return self.expiration

    def get_message_id(self):
        """
        Internal method for getting the message id.
        """
        return self.msgid

    def get_priority(self):
        """
        Internal method for getting the priority.
        """
        return self.priority

    def get_state(self):
        """
        Internal method for getting the message state.
        """
        return self.state

    def set_correlation(self, str value):
        """
        Internal method for setting the correlation.
        """
        self.correlation = value

    def set_delay(self, int32_t value):
        """
        Internal method for setting the delay.
        """
        self.delay = value

    def set_exception_queue(self, str value):
        """
        Internal method for setting the exception queue.
        """
        self.exceptionq = value

    def set_expiration(self, int32_t value):
        """
        Internal method for setting the message expiration.
        """
        self.expiration = value

    def set_payload_bytes(self, bytes value):
        """
        Internal method for setting the payload from bytes.
        """
        self.payload_obj = value

    def set_payload_object(self, ThinDbObjectImpl value):
        """
        Internal method for setting the payload from an object.
        """
        if not isinstance(value, ThinDbObjectImpl):
            raise TypeError("Expected ThinDbObjectImpl instance.")
        self.payload_obj = value

    def set_payload_json(self, object json_val):
        """
        Internal method for setting the payload from a JSON object
        """
        self.payload_obj = json_val

    def set_priority(self, int32_t value):
        """
        Internal method for setting the priority.
        """
        self.priority = value

    def set_recipients(self, list value):
        """
        Internal method for setting the recipients list.
        """
        self.recipients = value



#------------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# aq_base.pyx
#
# Cython file defining the base class for messages that are sent to the
# database and the responses that are received by the client for enqueing and
# dequeuing AQ messages (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class AqBaseMessage(Message):
    cdef:
        BaseThinQueueImpl queue_impl
        ThinDeqOptionsImpl deq_options_impl
        ThinEnqOptionsImpl enq_options_impl
        bint no_msg_found

    cdef object _process_date(self, ReadBuffer buf):
        """
        Processes a date found in the buffer.
        """
        cdef:
            const char_type *ptr
            uint32_t num_bytes
            OracleData data
            ssize_t length
        buf.read_ub4(&num_bytes)
        if num_bytes > 0:
            buf.read_raw_bytes_and_length(&ptr, &length)
            decode_date(ptr, length, &data.buffer)
            return convert_date_to_python(&data.buffer)

    cdef int _process_error_info(self, ReadBuffer buf) except -1:
        """
        Process error information from the buffer. If the error that indicates
        that no messages were received is detected, the error is cleared and
        the flag set so that the dequeue can handle that case.
        """
        Message._process_error_info(self, buf)
        if self.error_info.num == TNS_ERR_NO_MESSAGES_FOUND:
            self.error_info.num = 0
            self.error_occurred = False
            self.no_msg_found = True

    cdef int _process_extensions(self, ReadBuffer buf,
                                 ThinMsgPropsImpl props_impl) except -1:
        """
        Processes extensions to the message property object returned by the
        database.
        """
        cdef:
            bytes text_value, binary_value, value
            uint32_t i, num_extensions
            uint16_t keyword
        buf.read_ub4(&num_extensions)
        if num_extensions > 0:
            buf.skip_ub1()
            for i in range(num_extensions):
                text_value = buf.read_bytes_with_length()
                binary_value = buf.read_bytes_with_length()
                value = text_value or binary_value
                buf.read_ub2(&keyword)
                if value is not None:
                    if keyword == TNS_AQ_EXT_KEYWORD_AGENT_NAME:
                        props_impl.sender_agent_name = value
                    elif keyword == TNS_AQ_EXT_KEYWORD_AGENT_ADDRESS:
                        props_impl.sender_agent_address = value
                    elif keyword == TNS_AQ_EXT_KEYWORD_AGENT_PROTOCOL:
                        props_impl.sender_agent_protocol = value
                    elif keyword == TNS_AQ_EXT_KEYWORD_ORIGINAL_MSGID:
                        props_impl.original_msg_id = value

    cdef bytes _process_msg_id(self, ReadBuffer buf):
        """
        Reads a message id from the buffer and returns it.
        """
        cdef const char_type *ptr
        ptr = buf.read_raw_bytes(TNS_AQ_MESSAGE_ID_LENGTH)
        return ptr[:TNS_AQ_MESSAGE_ID_LENGTH]

    cdef int _process_msg_props(self, ReadBuffer buf,
                                ThinMsgPropsImpl props_impl) except -1:
        """
        Processes a message property object returned by the database.
        """
        cdef uint32_t temp32
        buf.read_sb4(&props_impl.priority)
        buf.read_sb4(&props_impl.delay)
        buf.read_sb4(&props_impl.expiration)
        props_impl.correlation = buf.read_str_with_length()
        buf.read_sb4(&props_impl.num_attempts)
        props_impl.exceptionq = buf.read_str_with_length()
        buf.read_sb4(&props_impl.state)
        props_impl.enq_time = self._process_date(buf)
        props_impl.enq_txn_id = buf.read_bytes_with_length()
        self._process_extensions(buf, props_impl)
        buf.read_ub4(&temp32)                       # user properties
        if temp32 > 0:
            errors._raise_err(errors.ERR_NOT_IMPLEMENTED)
        buf.skip_ub4()                              # csn
        buf.skip_ub4()                              # dsn
        buf.skip_ub4()                              # flags
        if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_21_1:
            buf.skip_ub4()                          # shard number

    cdef object _process_payload(self, ReadBuffer buf):
        """
        Processes the payload for an enqueued message returned by the database.
        """
        cdef:
            ThinDbObjectImpl obj_impl
            uint32_t image_length
            bytes payload
        if self.queue_impl.payload_type is not None:
            obj_impl = buf.read_dbobject(self.queue_impl.payload_type)
            if obj_impl is None:
                obj_impl = self.queue_impl.payload_type.create_new_object()
            return PY_TYPE_DB_OBJECT._from_impl(obj_impl)
        else:
            buf.read_bytes_with_length()            # TOID
            buf.read_bytes_with_length()            # OID
            buf.read_bytes_with_length()            # snapshot
            buf.skip_ub2()                          # version no
            buf.read_ub4(&image_length)             # image length
            buf.skip_ub2()                          # flags
            if image_length > 0:
                payload = buf.read_bytes()[4:image_length]
                if self.queue_impl.is_json:
                    return self.conn_impl.decode_oson(payload)
                return payload
            elif not self.queue_impl.is_json:
                return b''

    cdef object _process_recipients(self, ReadBuffer buf):
        """
        Process recipients for a message. Currently this is unsupported.
        """
        cdef uint32_t temp32
        buf.read_ub4(&temp32)
        if temp32 > 0:
            errors._raise_err(errors.ERR_NOT_IMPLEMENTED)
        return []

    cdef int _write_msg_props(self, WriteBuffer buf,
                              ThinMsgPropsImpl props_impl) except -1:
        """
        Write a message property object to the buffer.
        """
        buf.write_ub4(props_impl.priority)
        buf.write_ub4(props_impl.delay)
        buf.write_sb4(props_impl.expiration)
        self._write_value_with_length(buf, props_impl.correlation)
        buf.write_ub4(0)                            # number of attempts
        self._write_value_with_length(buf, props_impl.exceptionq)
        buf.write_ub4(props_impl.state)
        buf.write_ub4(0)                            # enqueue time length
        self._write_value_with_length(buf, props_impl.enq_txn_id)
        buf.write_ub4(4)                            # number of extensions
        buf.write_uint8(0x0e)                       # unknown extra byte
        buf.write_extension_values(None, None, TNS_AQ_EXT_KEYWORD_AGENT_NAME)
        buf.write_extension_values(None, None,
                                   TNS_AQ_EXT_KEYWORD_AGENT_ADDRESS)
        buf.write_extension_values(None, b'\x00',
                                   TNS_AQ_EXT_KEYWORD_AGENT_PROTOCOL)
        buf.write_extension_values(None, None,
                                   TNS_AQ_EXT_KEYWORD_ORIGINAL_MSGID)
        buf.write_ub4(0)                            # user property
        buf.write_ub4(0)                            # cscn
        buf.write_ub4(0)                            # dscn
        buf.write_ub4(0)                            # flags
        if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_21_1:
            buf.write_ub4(0xffffffffl)              # shard id

    cdef int _write_payload(self, WriteBuffer buf,
                            ThinMsgPropsImpl props_impl) except -1:
        """
        Writes the payload of the message property object to the buffer.
        """
        if self.queue_impl.is_json:
            buf.write_oson(props_impl.payload_obj,
                           self.conn_impl._oson_max_fname_size, False)
        elif self.queue_impl.payload_type is not None:
            buf.write_dbobject(props_impl.payload_obj)
        else:
            buf.write_bytes(props_impl.payload_obj)

    cdef int _write_value_with_length(self, WriteBuffer buf,
                                      object value) except -1:
        """
        Write a string to the buffer, prefixed by a length.
        """
        cdef bytes value_bytes
        if value is None:
            buf.write_ub4(0)
        else:
            if isinstance(value, str):
                value_bytes = value.encode()
            else:
                value_bytes = value
            buf.write_ub4(len(value_bytes))
            buf.write_bytes_with_length(value_bytes)



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E1200 - Module for testing the use of the statement cache
"""

import random

import test_env


class TestCase(test_env.BaseTestCase):
    requires_connection = False

    def test_ext_1200(self):
        "E1200 - verify statement cache is used"
        statements = [
            "select 0 from dual",
            "select 1 from dual",
            "select 2 from dual",
        ]
        with test_env.get_connection(stmtcachesize=len(statements)) as conn:
            self.setup_parse_count_checker(conn)
            for i in range(500):
                sql = random.choice(statements)
                with conn.cursor() as cursor:
                    cursor.execute(sql)
                    cursor.fetchall()
            self.assertParseCount(3)

    def test_ext_1201(self):
        "E1201 - verify statement cache uses LIFO"
        statements = [
            "select 0 from dual",
            "select 1 from dual",
            "select 2 from dual",
        ]
        with test_env.get_connection(stmtcachesize=len(statements)) as conn:
            self.setup_parse_count_checker(conn)
            for sql in statements:
                with conn.cursor() as cursor:
                    cursor.execute(sql)
                    cursor.fetchall()
            with conn.cursor() as cursor:
                cursor.execute("begin null; end;")
            self.assertParseCount(4)
            for sql in statements[1:]:
                with conn.cursor() as cursor:
                    cursor.execute(sql)
                    cursor.fetchall()
            self.assertParseCount(0)
            with conn.cursor() as cursor:
                cursor.execute(statements[0])
                cursor.fetchall()
            self.assertParseCount(1)

    def test_ext_1202(self):
        "E1202 - verify copied statement is independent of cached statement"
        sql = "select : val from dual"
        value1 = "One"
        value2 = "Four"
        value3 = "Five"
        with test_env.get_connection() as conn:
            self.setup_parse_count_checker(conn)
            with conn.cursor() as cursor:
                cursor.execute(sql, [value1])
                rows = cursor.fetchall()
                self.assertEqual(rows, [(value1,)])
            self.assertParseCount(1)
            with conn.cursor() as cursor:
                cursor.execute(sql, [value1])
                self.assertParseCount(0)
                with conn.cursor() as copyCursor:
                    copyCursor.execute(sql, [value2])
                    rows = copyCursor.fetchall()
                    self.assertEqual(rows, [(value2,)])
                rows = cursor.fetchall()
                self.assertEqual(rows, [(value1,)])
            self.assertParseCount(1)
            with conn.cursor() as cursor:
                cursor.execute(sql, [value3])
                rows = cursor.fetchall()
                self.assertEqual(rows, [(value3,)])
            self.assertParseCount(0)


if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2022, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# utils.pyx
#
# Cython file defining utility methods (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

cdef int _set_app_context_param(dict args, str name, object target) except -1:
    """
    Sets an application context parameter to the value provided in the
    dictionary, if a value is provided. This value is then set directly on the
    target.
    """
    in_val = args.get(name)
    if in_val is not None:
        message = (
            "appcontext should be a list of 3-tuples and each 3-tuple should "
            "contain three strings"
        )
        if not isinstance(in_val, list):
            raise TypeError(message)
        for entry in in_val:
            if not isinstance(entry, tuple) or len(entry) != 3:
                raise TypeError(message)
            for value in entry:
                if not isinstance(value, str):
                    raise TypeError(message)
        setattr(target, name, in_val)


cdef int _set_bool_param(dict args, str name, bint *out_val) except -1:
    """
    Sets a boolean parameter to the value provided in the dictionary. This can
    be a case-insenstive string matching on/off, yes/no or true/false (such as
    when parsed from a connect string). It can also be a directly passed
    argument which will be explicitly converted to a boolean value.
    """
    in_val = args.get(name)
    if in_val is not None:
        if isinstance(in_val, str):
            out_val[0] = (in_val.strip().lower() in ("on", "yes", "true"))
        else:
            out_val[0] = bool(in_val)


cdef int _set_duration_param(dict args, str name, double *out_val) except -1:
    """
    Sets a duration parameter to the value provided in the dictionary. This can
    be a string (such as when parsed from a connect string) containing a
    floating point value followed by an otional unit: ms (milliseconds), sec
    (seconds) or min (minutes). It can also be a directly passed argument which
    will be explicitly converted to a floating point value.
    """
    in_val = args.get(name)
    if in_val is not None:
        if isinstance(in_val, str):
            in_val = in_val.strip().lower()
            if in_val.endswith("sec"):
                out_val[0] = float(in_val[:-3].strip())
            elif in_val.endswith("ms"):
                out_val[0] = float(in_val[:-2].strip()) / 1000
            elif in_val.endswith("min"):
                out_val[0] = float(in_val[:-3].strip()) * 60
            else:
                out_val[0] = float(in_val.strip())
        else:
            out_val[0] = float(in_val)


cdef int _set_enum_param(dict args, str name, object enum_obj,
                         uint32_t* out_val) except -1:
    """
    Sets an unsigned integer parameter to the value provided in the dictionary.
    If digits are not provided, the value is looked up in the provided
    enumeration.
    """
    in_val = args.get(name)
    if in_val is not None:
        if not isinstance(in_val, str) or in_val.isdigit():
            out_val[0] = int(in_val)
            if isinstance(in_val, enum_obj):
                return 0
            try:
                enum_obj(out_val[0])
                return 0
            except ValueError:
                pass
        else:
            enum_val = getattr(enum_obj, in_val.upper(), None)
            if enum_val is not None:
                out_val[0] = enum_val.value
                return 0
        errors._raise_err(errors.ERR_INVALID_ENUM_VALUE,
                          name=enum_obj.__name__, value=in_val)


cdef int _set_int_param(dict args, str name, int* out_val) except -1:
    """
    Sets an integer parameter to the value provided in the dictionary. This
    can be a string (such as when parsed from a connect string). It can also be
    a directly passed argument which will be explicitly converted to an integer
    value.
    """
    in_val = args.get(name)
    if in_val is not None:
        out_val[0] = int(in_val)


cdef int _set_uint_param(dict args, str name, uint32_t* out_val) except -1:
    """
    Sets an unsigned integer parameter to the value provided in the dictionary.
    This can be a string (such as when parsed from a connect string). It can
    also be a directly passed argument which will be explicitly converted to an
    integer value.
    """
    in_val = args.get(name)
    if in_val is not None:
        out_val[0] = int(in_val)


cdef int _set_obj_param(dict args, str name, object target) except -1:
    """
    Sets an object parameter to the value provided in the dictionary, if a
    value is provided. This value is then set directly on the target.
    """
    in_val = args.get(name)
    if in_val is not None:
        setattr(target, name, in_val)


cdef int _set_ssl_version_param(dict args, str name, object target) except -1:
    """
    Sets a SSL version parameter to the value specified. This must be one of
    the values "tlsv1.2" or "tlsv1.3". If it is not one of these values
    an error is raised. If a value is specified and meets the criteria it is
    set directly on the target (since strings are treated as Python objects).
    """
    in_val = args.get(name)
    if in_val is not None:
        if isinstance(in_val, str):
            in_val = in_val.lower()
            if in_val == "tlsv1.2":
                in_val = ssl.TLSVersion.TLSv1_2
            elif in_val == "tlsv1.3":
                in_val = ssl.TLSVersion.TLSv1_3
        if in_val is not ssl.TLSVersion.TLSv1_2 \
                and in_val is not ssl.TLSVersion.TLSv1_3:
            errors._raise_err(errors.ERR_INVALID_SSL_VERSION,
                              ssl_version=in_val)
        setattr(target, name, in_val)


cdef int _set_str_param(dict args, str name, object target, bint check_network_character_set = False) except -1:
    """
    Sets a string parameter to the value provided in the dictionary. If a value
    is specified it is set directly on the target (since strings are treated as
    Python objects). Note that if check_network_character_set is True, an
    exception is thrown if the sanitized value does not match the input value.
    """
    in_val = args.get(name)
    if in_val:
        in_val = str(in_val)
        if check_network_character_set:
            if sanitize(in_val) != in_val:
                errors._raise_err(errors.ERR_INVALID_NETWORK_NAME, name=name)
        setattr(target, name, in_val)


def get_array_type_code_uint32():
    """
    Returns the type code to use for array.array that will store uint32_t.
    """
    cdef:
        array.array temp_array
        str type_code
    global ARRAY_TYPE_CODE_UINT32
    if ARRAY_TYPE_CODE_UINT32 is None:
        for type_code in ("I", "L"):
            temp_array = array.array(type_code)
            if temp_array.itemsize == 4:
                ARRAY_TYPE_CODE_UINT32 = type_code
                break
    return ARRAY_TYPE_CODE_UINT32


def init_base_impl(package):
    """
    Initializes globals after the package has been completely initialized. This
    is to avoid circular imports and eliminate the need for global lookups.
    """
    global \
        errors, \
        exceptions, \
        utils, \
        DRIVER_VERSION, \
        ENUM_AUTH_MODE, \
        ENUM_POOL_GET_MODE, \
        ENUM_PURITY, \
        PY_TYPE_ASYNC_CURSOR, \
        PY_TYPE_ASYNC_LOB, \
        PY_TYPE_CONNECT_PARAMS, \
        PY_TYPE_CURSOR, \
        PY_TYPE_DATAFRAME, \
        PY_TYPE_DB_OBJECT, \
        PY_TYPE_DB_OBJECT_TYPE, \
        PY_TYPE_FETCHINFO, \
        PY_TYPE_INTERVAL_YM, \
        PY_TYPE_JSON_ID, \
        PY_TYPE_LOB, \
        PY_TYPE_MESSAGE, \
        PY_TYPE_MESSAGE_QUERY, \
        PY_TYPE_MESSAGE_ROW, \
        PY_TYPE_MESSAGE_TABLE, \
        PY_TYPE_POOL_PARAMS, \
        PY_TYPE_SPARSE_VECTOR, \
        PY_TYPE_VAR

    errors = package.errors
    exceptions = package.exceptions
    utils = package.utils
    DRIVER_VERSION = package.__version__
    ENUM_AUTH_MODE = package.AuthMode
    ENUM_PURITY = package.Purity
    ENUM_POOL_GET_MODE = package.PoolGetMode
    PY_TYPE_ASYNC_CURSOR = package.AsyncCursor
    PY_TYPE_ASYNC_LOB = package.AsyncLOB
    PY_TYPE_CONNECT_PARAMS = package.ConnectParams
    PY_TYPE_CURSOR = package.Cursor
    PY_TYPE_DATAFRAME = package.OracleDataFrame
    PY_TYPE_DB_OBJECT = package.DbObject
    PY_TYPE_DB_OBJECT_TYPE = package.DbObjectType
    PY_TYPE_FETCHINFO = package.FetchInfo
    PY_TYPE_INTERVAL_YM = package.IntervalYM
    PY_TYPE_JSON_ID = package.JsonId
    PY_TYPE_LOB = package.LOB
    PY_TYPE_MESSAGE = package.Message
    PY_TYPE_MESSAGE_QUERY = package.MessageQuery
    PY_TYPE_MESSAGE_ROW = package.MessageRow
    PY_TYPE_MESSAGE_TABLE = package.MessageTable
    PY_TYPE_POOL_PARAMS = package.PoolParams
    PY_TYPE_SPARSE_VECTOR = package.SparseVector
    PY_TYPE_VAR = package.Var


def sanitize(str value):
    """
    Replaces invalid characters in a string with characters guaranteed to be in
    the network character set.
    """
    cdef str sanitized_value = value

    # strip single or double quotes from the start and end
    if sanitized_value.startswith(("'", '"')):
        sanitized_value = sanitized_value[1:]
    if sanitized_value.endswith(("'", '"')):
        sanitized_value = sanitized_value[:-1]

    # replace invalid characters with '?'
    sanitized_value = "".join(
        char if char in VALID_NETWORK_NAME_CHARS else "?"
        for char in sanitized_value
    )

    # if the last character is a backslash, replace it with '?'
    if sanitized_value.endswith("\\"):
        sanitized_value = sanitized_value[:-1] + "?"

    return sanitized_value



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# tpc_change_state.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for changing the state of a two
# phase commit (TPC) transaction (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class TransactionChangeStateMessage(Message):
    """
    Used for two-phase commit (TPC) transaction change state: commit, rollback,
    forget, etc.
    """
    cdef:
        uint32_t operation, state, flags
        bytes context
        object xid

    cdef int _initialize_hook(self) except -1:
        """
        Perform initialization.
        """
        self.function_code = TNS_FUNC_TPC_TXN_CHANGE_STATE

    cdef int _process_return_parameters(self, ReadBuffer buf) except -1:
        """
        Process the parameters returned by the database.
        """
        buf.read_ub4(&self.state)

    cdef int _write_message(self, WriteBuffer buf) except -1:
        """
        Writes the message to the database.
        """
        cdef:
            bytes global_transaction_id, branch_qualifier, xid_bytes
            uint32_t format_id = 0

        # acquire data to send to the server
        if self.xid is not None:
            format_id = self.xid[0]
            global_transaction_id = self.xid[1] \
                    if isinstance(self.xid[1], bytes) \
                    else self.xid[1].encode()
            branch_qualifier = self.xid[2] \
                    if isinstance(self.xid[2], bytes) \
                    else self.xid[2].encode()
            xid_bytes = global_transaction_id + branch_qualifier
            xid_bytes += bytes(128 - len(xid_bytes))

        self._write_function_code(buf)
        buf.write_ub4(self.operation)
        if self.context is not None:
            buf.write_uint8(1)              # pointer (context)
            buf.write_ub4(len(self.context))
        else:
            buf.write_uint8(0)              # pointer (context)
            buf.write_ub4(0)                # context length
        if self.xid is not None:
            buf.write_ub4(format_id)
            buf.write_ub4(len(global_transaction_id))
            buf.write_ub4(len(branch_qualifier))
            buf.write_uint8(1)              # pointer (xid)
            buf.write_ub4(len(xid_bytes))
        else:
            buf.write_ub4(0)                # format id
            buf.write_ub4(0)                # global transaction id length
            buf.write_ub4(0)                # branch qualifier length
            buf.write_uint8(0)              # pointer (xid)
            buf.write_ub4(0)                # XID length
        buf.write_ub4(0)                    # timeout
        buf.write_ub4(self.state)
        buf.write_uint8(1)                  # pointer (out state)
        buf.write_ub4(self.flags)
        if self.context is not None:
            buf.write_bytes(self.context)
        if self.xid is not None:
            buf.write_bytes(xid_bytes)



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# crypto.pyx
#
# Cython file defining the cryptographic methods used by the thin client
# (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

try:
    from cryptography import x509
    from cryptography.hazmat.primitives import hashes, serialization
    from cryptography.hazmat.primitives.ciphers import algorithms, modes, Cipher
    from cryptography.hazmat.primitives.asymmetric import padding
    from cryptography.hazmat.primitives.kdf import pbkdf2
except Exception as e:
    CRYPTOGRAPHY_IMPORT_ERROR = e


DN_REGEX = '(?:^|,\s?)(?:(?P<name>[A-Z]+)=(?P<val>"(?:[^"]|"")+"|[^,]+))+'
PEM_WALLET_FILE_NAME = "ewallet.pem"

def _name_matches(name_to_check, cert_name):
    """
    Returns a boolean indicating if the name to check matches with the
    certificate name. The certificate name may contain a wildcard (*)
    character.
    """

    # check for a full match (case insensitive)
    cert_name = cert_name.lower()
    name_to_check = name_to_check.lower()
    if name_to_check == cert_name:
        return True

    # ensure that more than one label exists in both the name to check and the
    # certificate name
    check_pos = name_to_check.find(".")
    cert_pos = cert_name.find(".")
    if check_pos <= 0 or cert_pos <= 0:
        return False

    # ensure that the right hand labels all match
    if name_to_check[check_pos:] != cert_name[cert_pos:]:
        return False

    # match wildcards, if applicable
    cert_label = cert_name[:cert_pos]
    check_label = name_to_check[:check_pos]
    if cert_label == "*":
        return True
    elif cert_label.startswith("*"):
        return check_label.endswith(cert_label[1:])
    elif cert_label.endswith("*"):
        return check_label.startswith(cert_label[:-1])
    wildcard_pos = cert_name.find("*")
    if wildcard_pos < 0:
        return False
    return check_label.startswith(cert_label[:wildcard_pos]) \
            and check_label.endswith(cert_label[wildcard_pos + 1:])


def check_server_dn(sock, expected_dn, expected_name):
    """
    Validates the server distinguished name (if one is specified) or the
    simple name (if a distinguished name is not present).
    """
    cert_data = sock.getpeercert(binary_form=True)
    cert = x509.load_der_x509_certificate(cert_data)
    if expected_dn is not None:
        server_dn = cert.subject.rfc4514_string()
        expected_dn_dict = dict(re.findall(DN_REGEX, expected_dn))
        server_dn_dict = dict(re.findall(DN_REGEX, server_dn))
        if server_dn_dict != expected_dn_dict:
            errors._raise_err(errors.ERR_INVALID_SERVER_CERT_DN,
                              expected_dn=expected_dn)
    else:
        try:
            ext = cert.extensions.get_extension_for_oid(
                x509.oid.ExtensionOID.SUBJECT_ALTERNATIVE_NAME
            )
            for name in ext.value.get_values_for_type(x509.DNSName):
                if _name_matches(expected_name, name):
                    return
        except x509.ExtensionNotFound:
            pass
        for name in cert.subject.get_attributes_for_oid(
            x509.oid.NameOID.COMMON_NAME
        ):
            if _name_matches(expected_name, name.value):
                return
        errors._raise_err(errors.ERR_INVALID_SERVER_NAME,
                          expected_name=expected_name)


def decrypt_cbc(key, encrypted_text):
    """
    Decrypt the given text using the given key.
    """
    iv = bytes(16)
    algo = algorithms.AES(key)
    cipher = Cipher(algo, modes.CBC(iv))
    decryptor = cipher.decryptor()
    return decryptor.update(encrypted_text)


def encrypt_cbc(key, plain_text, zeros=False):
    """
    Encrypt the given text using the given key. If the zeros flag is set, use
    zero padding if required. Otherwise, use number padding.
    """
    block_size = 16
    iv = bytes(block_size)
    algo = algorithms.AES(key)
    cipher = Cipher(algo, modes.CBC(iv))
    encryptor = cipher.encryptor()
    n = block_size - len(plain_text) % block_size
    if n:
        if zeros:
            plain_text += bytes(n)
        else:
            plain_text += (bytes([n]) * n)
    return encryptor.update(plain_text) + encryptor.finalize()


def get_derived_key(key, salt, length, iterations):
    """
    Return a derived key using PBKDF2.
    """
    kdf = pbkdf2.PBKDF2HMAC(algorithm=hashes.SHA512(), salt=salt,
                            length=length, iterations=iterations)
    return kdf.derive(key)


def get_signature(private_key_str, text):
    """
    Returns a signed version of the given text (used for IAM token
    authentication) in base64 encoding.
    """
    private_key = serialization.load_pem_private_key(private_key_str.encode(),
                                                     password=None)
    sig = private_key.sign(text.encode(), padding.PKCS1v15(), hashes.SHA256())
    return base64.b64encode(sig).decode()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# pool.pyx
#
# Cython file defining the pool implementation class (embedded in
# thin_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseThinPoolImpl(BasePoolImpl):

    cdef:
        list _free_new_conn_impls
        list _free_used_conn_impls
        list _busy_conn_impls
        list _conn_impls_to_drop
        list _requests
        uint32_t _getmode
        uint32_t _stmt_cache_size
        uint32_t _timeout
        uint32_t _max_lifetime_session
        uint32_t _auth_mode
        uint32_t _open_count
        uint32_t _num_to_create
        int _ping_interval
        uint32_t _ping_timeout
        object _wait_timeout
        object _bg_task
        object _bg_task_condition
        object _condition
        object _timeout_task
        object _ssl_session
        bint _force_get
        bint _open

    def __init__(self, str dsn, PoolParamsImpl params):
        _check_cryptography()
        params._check_credentials()
        self.connect_params = params
        self.username = params.user
        self.dsn = dsn
        self.min = params.min
        self.max = params.max
        self.increment = params.increment
        self.homogeneous = params.homogeneous
        self.set_getmode(params.getmode)
        self.set_wait_timeout(params.wait_timeout)
        self.set_timeout(params.timeout)
        self._stmt_cache_size = params.stmtcachesize
        self._max_lifetime_session = params.max_lifetime_session
        self._ping_interval = params.ping_interval
        self._ping_timeout = params.ping_timeout
        self._free_new_conn_impls = []
        self._free_used_conn_impls = []
        self._busy_conn_impls = []
        self._conn_impls_to_drop = []
        self._requests = []
        self._num_to_create = self.min
        self._auth_mode = AUTH_MODE_DEFAULT
        self._open = True

    cdef int _add_request(self, PooledConnRequest request) except -1:
        """
        Adds a request for the background task to process.
        """
        request.bg_processing = True
        request.completed = False
        self._requests.append(request)
        self._notify_bg_task()

    cdef int _check_timeout(self) except -1:
        """
        Checks whether a timeout is in effect and that the number of
        connections exceeds the minimum and if so, starts a timer to check
        if these connections have expired. Only one task is ever started.
        The timeout value is increased by a second to allow for small time
        discrepancies.
        """
        if self._open and self._timeout_task is None and self._timeout > 0 \
                and self._open_count > self.min:
            self._start_timeout_task()

    cdef int _close_helper(self, bint force) except -1:
        """
        Helper function that closes all of the connections in the pool.
        """
        cdef BaseThinConnImpl conn_impl

        # if force parameter is not True and busy connections exist in the
        # pool or there are outstanding requests, raise an exception
        if not force and (self.get_busy_count() > 0 or self._requests):
            errors._raise_err(errors.ERR_POOL_HAS_BUSY_CONNECTIONS)

        # close all connections in the pool; this is done by simply adding
        # to the list of connections that require closing and then notifying
        # the background task to perform the work
        self._open = False
        for lst in (self._free_used_conn_impls,
                    self._free_new_conn_impls,
                    self._busy_conn_impls):
            self._conn_impls_to_drop.extend(lst)
            for conn_impl in lst:
                conn_impl._pool = None
            lst.clear()
        self._notify_bg_task()

    cdef PooledConnRequest _create_request(self, ConnectParamsImpl params):
        """
        Returns a poooled connection request suitable for establishing a
        connection to the pool with the given parameters.
        """
        cdef:
            ConnectParamsImpl creation_params = self.connect_params
            str pool_cclass = creation_params._default_description.cclass
            PooledConnRequest request
        request = PooledConnRequest.__new__(PooledConnRequest)
        request.pool_impl = self
        request.params = params
        request.cclass = params._default_description.cclass
        request.wants_new = (params._default_description.purity == PURITY_NEW)
        request.cclass_matches = \
                (request.cclass is None or request.cclass == pool_cclass)
        request.waiting = True
        return request

    cdef int _drop_conn_impl(self, BaseThinConnImpl conn_impl) except -1:
        """
        Helper method which adds a connection to the list of connections to be
        closed and notifies the background task.
        """
        conn_impl._pool = None
        if conn_impl._protocol._transport is not None:
            self._conn_impls_to_drop.append(conn_impl)
            self._notify_bg_task()
        self._ensure_min_connections()

    cdef int _drop_conn_impls_helper(self, list conn_impls_to_drop) except -1:
        """
        Helper method which drops the requested list of connections. Exceptions
        that take place while attempting to close the connection are ignored.
        """
        cdef BaseThinConnImpl conn_impl
        for conn_impl in conn_impls_to_drop:
            try:
                conn_impl._force_close()
            except:
                pass

    cdef int _ensure_min_connections(self) except -1:
        """
        Ensure that the minimum number of connections in the pool is
        maintained.
        """
        if self._open_count < self.min:
            self._num_to_create = max(self._num_to_create,
                                      self.min - self._open_count)
            self._notify_bg_task()

    cdef PooledConnRequest _get_next_request(self):
        """
        Get the next request to process.
        """
        cdef PooledConnRequest request
        for request in self._requests:
            if not request.waiting \
                    or request.requires_ping \
                    or request.is_replacing \
                    or request.is_extra \
                    or self._open_count < self.max:
                request.in_progress = request.waiting
                return request
            break

    cdef BaseThinConnImpl _post_acquire(self, BaseThinConnImpl conn_impl):
        """
        Called after an acquire has succeeded. The connection is added to the
        list of busy connections and is marked as being in a request.
        """
        self._busy_conn_impls.append(conn_impl)
        if conn_impl._protocol._caps.supports_request_boundaries:
            conn_impl._session_state_desired = TNS_SESSION_STATE_REQUEST_BEGIN
            conn_impl._in_request = True
        return conn_impl

    cdef int _post_create_conn_impl(self,
                                    BaseThinConnImpl conn_impl) except -1:
        """
        Called after a connection has been created without an associated
        request.
        """
        cdef PooledConnRequest request
        if conn_impl is None:
            self._num_to_create = 0
        elif not self._open:
            conn_impl._force_close()
        else:
            self._open_count += 1
            if self._num_to_create > 0:
                self._num_to_create -= 1
            for request in self._requests:
                if request.in_progress or request.conn_impl is not None \
                        or not request.waiting:
                    continue
                if request.cclass is None \
                        or request.cclass == conn_impl._cclass:
                    request.conn_impl = conn_impl
                    request.completed = True
                    self._requests.remove(request)
                    self._condition.notify_all()
                    break
                elif not request.cclass_matches \
                        and self._open_count >= self.max:
                    request.conn_impl = conn_impl
                    request.is_replacing = True
                    break
            else:
                self._free_new_conn_impls.append(conn_impl)
            self._check_timeout()

    cdef int _post_process_request(self, PooledConnRequest request) except -1:
        """
        Called after the request has been processed. This removes the request
        from the list of requests and adds the connection to the appropriate
        list depending on whether the waiter is still waiting for the request
        to be satisfied!
        """
        request.in_progress = False
        request.bg_processing = False
        if request.conn_impl is not None:
            request.completed = True
            if not request.is_replacing and not request.requires_ping:
                self._open_count += 1
                if self._num_to_create > 0:
                    self._num_to_create -= 1
            if not request.waiting:
                request.reject()
        elif request.requires_ping:
            self._open_count -= 1
            if self._num_to_create == 0 and self._open_count < self.min:
                self._num_to_create = self.min - self._open_count
        self._requests.remove(request)
        self._condition.notify_all()

    cdef int _pre_connect(self, BaseThinConnImpl conn_impl,
                          ConnectParamsImpl params) except -1:
        """
        Called before the connection is connected. The connection class and
        pool attributes are updated and the TLS session is stored on the
        transport for reuse. The timestamps are also retained for later use.
        """
        if params is not None:
            conn_impl._cclass = params._default_description.cclass
        else:
            conn_impl._cclass = self.connect_params._default_description.cclass
        conn_impl._pool = self
        conn_impl._time_created = time.monotonic()
        conn_impl._time_returned = conn_impl._time_created

    def _process_timeout(self):
        """
        Processes the timeout after the timer task completes. Drops any free
        connections that have expired (while maintaining the minimum number of
        connections in the pool).
        """
        self._timeout_task = None
        self._timeout_helper(self._free_new_conn_impls)
        self._timeout_helper(self._free_used_conn_impls)
        self._check_timeout()

    cdef int _return_connection_helper(self,
                                       BaseThinConnImpl conn_impl) except -1:
        """
        Returns the connection to the pool. If the connection was closed for
        some reason it will be dropped; otherwise, it will be returned to the
        list of connections available for further use. If an "extra" connection
        was created (because the pool has a mode of "force" get or because a
        different connection class than that used by the pool was requested)
        then it will be added to the pool or will replace an unused new
        connection or will be discarded depending on the current pool size.
        """
        cdef:
            bint is_open = conn_impl._protocol._transport is not None
            BaseThinDbObjectTypeCache type_cache
            PooledConnRequest request
            double tstamp
            int cache_num
        self._busy_conn_impls.remove(conn_impl)
        if conn_impl._dbobject_type_cache_num > 0:
            cache_num = conn_impl._dbobject_type_cache_num
            type_cache = get_dbobject_type_cache(cache_num)
            type_cache._clear_cursors()
        if not is_open:
            self._open_count -= 1
            self._ensure_min_connections()
        if conn_impl._is_pool_extra:
            conn_impl._is_pool_extra = False
            if is_open and self._open_count >= self.max:
                if self._free_new_conn_impls and self._open_count == self.max:
                    self._drop_conn_impl(self._free_new_conn_impls.pop(0))
                else:
                    self._open_count -= 1
                    self._drop_conn_impl(conn_impl)
                    is_open = False
        if is_open:
            conn_impl.warning = None
            conn_impl._time_returned = time.monotonic()
            if self._max_lifetime_session != 0:
                tstamp = conn_impl._time_created + self._max_lifetime_session
                if conn_impl._time_returned > tstamp:
                    self._open_count -= 1
                    self._drop_conn_impl(conn_impl)
                    is_open = False
        if is_open:
            for request in self._requests:
                if request.in_progress or request.wants_new \
                        or request.conn_impl is not None \
                        or not request.waiting:
                    continue
                if request.cclass is None \
                        or request.conn_impl._cclass == self.cclass:
                    request.conn_impl = conn_impl
                    request.completed = True
                    self._requests.remove(request)
                    self._condition.notify_all()
                    return 0
            self._free_used_conn_impls.append(conn_impl)
        self._check_timeout()

    cdef int _shutdown(self) except -1:
        """
        Called when the main interpreter has completed and only shutdown code
        is being executed. All connections in the pool are marked as non-pooled
        and the pool itself terminated.
        """
        cdef BaseThinConnImpl conn_impl
        with self._condition:
            self._open = False
            for lst in (self._free_used_conn_impls,
                        self._free_new_conn_impls,
                        self._busy_conn_impls):
                for conn_impl in lst:
                    conn_impl._pool = None
                lst.clear()
            self._requests.clear()
            self._notify_bg_task()
        self._bg_task.join()

    cdef int _start_timeout_task(self) except -1:
        """
        Starts the task for checking timeouts (differs for sync and async).
        """
        pass

    cdef int _timeout_helper(self, list conn_impls_to_check) except -1:
        """
        Helper method which checks the list of connections to see if any
        connections have expired (while maintaining the minimum number of
        connections in the pool).
        """
        cdef BaseThinConnImpl conn_impl
        current_time = time.monotonic()
        while conn_impls_to_check and self._open_count > self.min:
            conn_impl = conn_impls_to_check[0]
            if current_time - conn_impl._time_returned < self._timeout:
                break
            conn_impls_to_check.pop(0)
            self._drop_conn_impl(conn_impl)
            self._open_count -= 1

    def get_busy_count(self):
        """
        Internal method for getting the number of busy connections in the pool.
        """
        return len(self._busy_conn_impls)

    def get_getmode(self):
        """
        Internal method for getting the method by which connections are
        acquired from the pool.
        """
        return self._getmode

    def get_max_lifetime_session(self):
        """
        Internal method for getting the maximum lifetime of each session.
        """
        return self._max_lifetime_session

    def get_open_count(self):
        """
        Internal method for getting the number of connections in the pool.
        """
        return self._open_count

    def get_ping_interval(self):
        """
        Internal method for getting the value of the pool-ping-interval.
        """
        return self._ping_interval

    def get_stmt_cache_size(self):
        """
        Internal method for getting the size of the statement cache.
        """
        return self._stmt_cache_size

    def get_timeout(self):
        """
        Internal method for getting the timeout for idle sessions.
        """
        return self._timeout

    def get_wait_timeout(self):
        """
        Internal method for getting the wait timeout for acquiring sessions.
        """
        if self._getmode == POOL_GETMODE_TIMEDWAIT:
            return self._wait_timeout
        return 0

    def set_getmode(self, uint32_t value):
        """
        Internal method for setting the method by which connections are
        acquired from the pool.
        """
        if self._getmode != value:
            self._getmode = value
            self._force_get = (self._getmode == POOL_GETMODE_FORCEGET)
            if self._getmode == POOL_GETMODE_TIMEDWAIT:
                self._wait_timeout = 0
            else:
                self._wait_timeout = None

    def set_max_lifetime_session(self, uint32_t value):
        """
        Internal method for setting the maximum lifetime of each session.
        """
        self._max_lifetime_session = value

    def set_ping_interval(self, int value):
        """
        Internal method for setting the value of the pool-ping-interval.
        """
        self._ping_interval = value

    def set_stmt_cache_size(self, uint32_t value):
        """
        Internal method for setting the size of the statement cache.
        """
        self._stmt_cache_size = value

    def set_timeout(self, uint32_t value):
        """
        Internal method for setting the timeout for idle sessions.
        """
        self._timeout = value

    def set_wait_timeout(self, uint32_t value):
        """
        Internal method for setting the wait timeout for acquiring sessions.
        """
        if self._getmode == POOL_GETMODE_TIMEDWAIT:
            self._wait_timeout = value / 1000
        else:
            self._wait_timeout = None


cdef class ThinPoolImpl(BaseThinPoolImpl):

    def __init__(self, str dsn, PoolParamsImpl params):
        super().__init__(dsn, params)
        self._condition = threading.Condition()
        self._bg_task_condition = threading.Condition()
        self._bg_task = threading.Thread(target=self._bg_task_func)
        self._bg_task.start()

    def _bg_task_func(self):
        """
        Method which runs in a dedicated thread and is used to create
        connections and close them when needed. When first started, it creates
        pool.min connections. After that, it creates pool.increment connections
        up to the value of pool.max when needed and destroys connections when
        needed. It also pings connections when requested to do so. The thread
        terminates automatically when the pool is closed.
        """
        cdef:
            PooledConnRequest request = None
            BaseThinConnImpl conn_impl
            list conn_impls_to_drop
            uint32_t num_to_create

        # add to the list of pools that require closing
        pool_closer.add_pool(self)

        # perform task until pool is closed
        while self._open or self._conn_impls_to_drop:

            # check to see if there a request to process
            if request is None and self._open:
                with self._condition:
                    request = self._get_next_request()
            if request is not None and self._open:
                self._process_request(request)
                with self._condition:
                    self._post_process_request(request)
                    request = self._get_next_request()
                    continue

            # check to see if there is a connection that needs to be built
            with self._condition:
                num_to_create = self._num_to_create
            if num_to_create > 0 and self._open:
                try:
                    conn_impl = self._create_conn_impl()
                except:
                    conn_impl = None
                with self._condition:
                    self._post_create_conn_impl(conn_impl)
                    continue

            # check to see if there are any connections to drop
            with self._condition:
                conn_impls_to_drop = self._conn_impls_to_drop
                self._conn_impls_to_drop = []
            if conn_impls_to_drop:
                self._drop_conn_impls_helper(conn_impls_to_drop)
                continue

            # otherwise, nothing to do yet, wait for notifications!
            with self._bg_task_condition:
                self._bg_task_condition.wait()

        # stop the timeout task, if one is active
        if self._timeout_task is not None:
            self._timeout_task.cancel()

        # remove from the list of pools that require closing
        pool_closer.remove_pool(self)

    cdef ThinConnImpl _create_conn_impl(self, ConnectParamsImpl params=None):
        """
        Create a single connection using the pool's information. This
        connection may be placed in the pool or may be returned directly (such
        as when the pool is full and POOL_GETMODE_FORCEGET is being used).
        """
        cdef ThinConnImpl conn_impl
        conn_impl = ThinConnImpl(self.dsn, self.connect_params)
        self._pre_connect(conn_impl, params)
        conn_impl.connect(self.connect_params)
        return conn_impl

    def _notify_bg_task(self):
        """
        Notify the background task that work needs to be done.
        """
        with self._bg_task_condition:
            self._bg_task_condition.notify()

    cdef int _process_request(self, PooledConnRequest request) except -1:
        """
        Processes a request.
        """
        cdef BaseThinConnImpl conn_impl
        try:
            if request.requires_ping:
                try:
                    request.conn_impl.set_call_timeout(self._ping_timeout)
                    request.conn_impl.ping()
                    request.conn_impl.set_call_timeout(0)
                except exceptions.Error:
                    request.conn_impl._force_close()
                    request.conn_impl = None
            else:
                conn_impl = self._create_conn_impl(request.params)
                if request.conn_impl is not None:
                    request.conn_impl._force_close()
                request.conn_impl = conn_impl
                request.conn_impl._is_pool_extra = request.is_extra
        except Exception as e:
            request.exception = e

    cdef int _return_connection(self, BaseThinConnImpl conn_impl) except -1:
        """
        Returns the connection to the pool.
        """
        with self._condition:
            self._return_connection_helper(conn_impl)

    cdef int _start_timeout_task(self) except -1:
        """
        Starts the task for checking timeouts. The timeout value is increased
        by a second to allow for small time discrepancies.
        """
        def handler():
            with self._condition:
                self._process_timeout()
        self._timeout_task = threading.Timer(self._timeout + 1, handler)
        self._timeout_task.start()

    def acquire(self, ConnectParamsImpl params):
        """
        Internal method for acquiring a connection from the pool.
        """
        cdef PooledConnRequest request

        # if pool is closed, raise an exception
        if not self._open:
            errors._raise_err(errors.ERR_POOL_NOT_OPEN)

        # session tagging has not been implemented yet
        if params.tag is not None:
            raise NotImplementedError("Tagging has not been implemented yet")

        # wait until an acceptable connection is found
        request = self._create_request(params)
        with self._condition:
            try:
                self._condition.wait_for(request.fulfill, self._wait_timeout)
            except:
                if not request.bg_processing:
                    request.reject()
                raise
            finally:
                request.waiting = False
            if not request.completed:
                errors._raise_err(errors.ERR_POOL_NO_CONNECTION_AVAILABLE)
            return self._post_acquire(request.conn_impl)

    def close(self, bint force):
        """
        Internal method for closing the pool. Note that the thread to destroy
        pools gracefully may have already run, so if the close has already
        happened, nothing more needs to be done!
        """
        if self._open:
            with self._condition:
                self._close_helper(force)
            self._bg_task.join()

    def drop(self, ThinConnImpl conn_impl):
        """
        Internal method for dropping a connection from the pool.
        """
        with self._condition:
            self._open_count -= 1
            self._busy_conn_impls.remove(conn_impl)
            self._drop_conn_impl(conn_impl)
            self._condition.notify()


cdef class AsyncThinPoolImpl(BaseThinPoolImpl):

    cdef:
        object _bg_notify_task

    def __init__(self, str dsn, PoolParamsImpl params):
        super().__init__(dsn, params)
        self._condition = asyncio.Condition()
        self._bg_task_condition = asyncio.Condition()
        self._bg_task = asyncio.create_task(self._bg_task_func())

    async def _acquire_helper(self, PooledConnRequest request):
        """
        Helper function for acquiring a connection from the pool.
        """
        async with self._condition:
            try:
                await self._condition.wait_for(request.fulfill)
            except:
                if not request.bg_processing:
                    request.reject()
                raise
            finally:
                request.waiting = False

    async def _bg_task_func(self):
        """
        Method which runs in a dedicated task and is used to create connections
        and close them when needed. When first started, it creates pool.min
        connections. After that, it creates pool.increment connections up to
        the value of pool.max when needed and destroys connections when needed.
        The task terminates automatically when the pool is closed.
        """
        cdef:
            PooledConnRequest request = None
            BaseThinConnImpl conn_impl
            list conn_impls_to_drop
            uint32_t num_to_create

        # perform task until pool is closed
        while self._open or self._conn_impls_to_drop:

            # check to see if there a request to process
            if request is None and self._open:
                async with self._condition:
                    request = self._get_next_request()
            if request is not None and self._open:
                await self._process_request(request)
                async with self._condition:
                    self._post_process_request(request)
                    request = self._get_next_request()
                    continue

            # check to see if there is a connection that needs to be built
            async with self._condition:
                num_to_create = self._num_to_create
            if num_to_create > 0 and self._open:
                try:
                    conn_impl = await self._create_conn_impl()
                except:
                    conn_impl = None
                async with self._condition:
                    self._post_create_conn_impl(conn_impl)
                    continue

            # check to see if there are any connections to drop
            async with self._condition:
                conn_impls_to_drop = self._conn_impls_to_drop
                self._conn_impls_to_drop = []
            if conn_impls_to_drop:
                self._drop_conn_impls_helper(conn_impls_to_drop)
                continue

            # otherwise, nothing to do yet, wait for notifications!
            async with self._bg_task_condition:
                await self._bg_task_condition.wait()

        # stop the timeout task, if one is active
        if self._timeout_task is not None:
            self._timeout_task.cancel()

    async def _create_conn_impl(self, ConnectParamsImpl params=None):
        """
        Create a single connection using the pool's information. This
        connection may be placed in the pool or may be returned directly (such
        as when the pool is full and POOL_GETMODE_FORCEGET is being used).
        """
        cdef AsyncThinConnImpl conn_impl
        conn_impl = AsyncThinConnImpl(self.dsn, self.connect_params)
        self._pre_connect(conn_impl, params)
        await conn_impl.connect(self.connect_params)
        return conn_impl

    def _notify_bg_task(self):
        """
        Notify the background task that work needs to be done.
        """
        if self._bg_notify_task is None or self._bg_notify_task.done():
            async def helper():
                async with self._bg_task_condition:
                    self._bg_task_condition.notify()
            self._bg_notify_task = asyncio.create_task(helper())

    async def _process_request(self, PooledConnRequest request):
        """
        Processes a request.
        """
        cdef BaseThinConnImpl conn_impl
        try:
            if request.requires_ping:
                try:
                    request.conn_impl.set_call_timeout(self._ping_timeout)
                    await request.conn_impl.ping()
                    request.conn_impl.set_call_timeout(0)
                except exceptions.Error:
                    request.conn_impl._force_close()
                    request.conn_impl = None
            else:
                conn_impl = await self._create_conn_impl(request.params)
                if request.conn_impl is not None:
                    request.conn_impl._force_close()
                request.conn_impl = conn_impl
                request.conn_impl._is_pool_extra = request.is_extra
        except Exception as e:
            request.exception = e

    async def _return_connection(self, BaseThinConnImpl conn_impl):
        """
        Returns the connection to the pool.
        """
        async with self._condition:
            self._return_connection_helper(conn_impl)

    cdef int _start_timeout_task(self) except -1:
        """
        Starts the task for checking timeouts. The timeout value is increased
        by a second to allow for small time discrepancies.
        """
        async def process_timeout():
            await asyncio.sleep(self._timeout + 1)
            async with self._condition:
                self._process_timeout()
        self._timeout_task = asyncio.create_task(process_timeout())

    async def acquire(self, ConnectParamsImpl params):
        """
        Internal method for acquiring a connection from the pool.
        """
        cdef PooledConnRequest request

        # if pool is closed, raise an exception
        if not self._open:
            errors._raise_err(errors.ERR_POOL_NOT_OPEN)

        # session tagging has not been implemented yet
        if params.tag is not None:
            raise NotImplementedError("Tagging has not been implemented yet")

        # use the helper function to allow for a timeout since asyncio
        # condition variables do not have that capability directly
        request = self._create_request(params)
        try:
            await asyncio.wait_for(
                self._acquire_helper(request), self._wait_timeout
            )
        except asyncio.TimeoutError:
            errors._raise_err(errors.ERR_POOL_NO_CONNECTION_AVAILABLE)
        return self._post_acquire(request.conn_impl)

    async def close(self, bint force):
        """
        Internal method for closing the pool.
        """
        async with self._condition:
            self._close_helper(force)
        await self._bg_task

    async def drop(self, AsyncThinConnImpl conn_impl):
        """
        Internal method for dropping a connection from the pool.
        """
        async with self._condition:
            self._open_count -= 1
            self._busy_conn_impls.remove(conn_impl)
            self._drop_conn_impl(conn_impl)
            self._condition.notify()


@cython.freelist(20)
cdef class PooledConnRequest:
    cdef:
        BaseThinPoolImpl pool_impl
        BaseThinConnImpl conn_impl
        ConnectParamsImpl params
        str cclass
        object exception
        bint cclass_matches
        bint requires_ping
        bint wants_new
        bint bg_processing
        bint is_extra
        bint is_replacing
        bint in_progress
        bint completed
        bint waiting

    cdef int _check_connection(self, BaseThinConnImpl conn_impl) except -1:
        """
        Checks the connection to see if it can be used. First, if any control
        packets are sent that indicate that the connection should be closed,
        the connection is indeed closed. After that, a flag is updated to the
        caller indicating that a ping is required according to the pool
        configuration.
        """
        cdef:
            ReadBuffer buf = conn_impl._protocol._read_buf
            double elapsed_time, min_create_time
            bint has_data_ready
        if not buf._transport._is_async:
            while buf._pending_error_num == 0:
                buf._transport.has_data_ready(&has_data_ready)
                if not has_data_ready:
                    break
                buf.check_control_packet()
        if buf._pending_error_num != 0:
            self.pool_impl._open_count -= 1
            self.pool_impl._drop_conn_impl(conn_impl)
            return 0
        elif self.pool_impl._max_lifetime_session > 0:
            min_create_time = \
                    time.monotonic() - self.pool_impl._max_lifetime_session
            if conn_impl._time_created < min_create_time:
                self.pool_impl._open_count -= 1
                self.pool_impl._drop_conn_impl(conn_impl)
                return 0
        self.conn_impl = conn_impl
        if self.pool_impl._ping_interval == 0:
            self.requires_ping = True
        elif self.pool_impl._ping_interval > 0:
            elapsed_time = time.monotonic() - conn_impl._time_returned
            if elapsed_time > self.pool_impl._ping_interval:
                self.requires_ping = True
        if self.requires_ping:
            self.pool_impl._add_request(self)
        else:
            self.completed = True

    def fulfill(self):
        """
        Fulfills the connection request. If a connection is available and does
        not require a ping to validate it, it is returned immediately. If a
        connection needs to be created or requires a ping to validate it, the
        background function is notified.
        """
        cdef:
            BaseThinPoolImpl pool = self.pool_impl
            BaseThinConnImpl conn_impl
            ssize_t ix
            object exc

        # if an exception was raised in the background thread, raise it now
        if self.exception is not None:
            raise self.exception

        # if the request is completed, waiting can end
        elif self.completed:
            return True

        # if the background task is still working on this request, go back to
        # waiting
        elif self.bg_processing:
            return False

        # check for an available used connection (only permitted if a new
        # connection is not required); in addition, ensure that the connection
        # class matches
        if not self.wants_new and pool._free_used_conn_impls:
            ix = len(pool._free_used_conn_impls) - 1
            for conn_impl in reversed(pool._free_used_conn_impls):
                if self.cclass is None or conn_impl._cclass == self.cclass:
                    pool._free_used_conn_impls.pop(ix)
                    self._check_connection(conn_impl)
                    if self.completed or self.requires_ping:
                        return self.completed
                ix -= 1

        # check for an available new connection (only permitted if the
        # connection class matches)
        if self.cclass_matches:
            while pool._free_new_conn_impls:
                conn_impl = pool._free_new_conn_impls.pop()
                self._check_connection(conn_impl)
                if self.completed or self.requires_ping:
                    return self.completed

        # no matching connections are available; if the pool is full, see if
        # any connections are available and if so, ask the background task to
        # create a new one to replace the non-matching one
        self.requires_ping = False
        if pool._open_count + pool._num_to_create >= pool.max:
            if pool._free_new_conn_impls:
                self.is_replacing = True
                conn_impl = pool._free_new_conn_impls.pop()
                pool._conn_impls_to_drop.append(conn_impl)
                pool._add_request(self)
                return False
            elif pool._free_used_conn_impls:
                self.is_replacing = True
                conn_impl = pool._free_used_conn_impls.pop()
                pool._conn_impls_to_drop.append(conn_impl)
                pool._add_request(self)
                return False
            elif pool._force_get:
                self.is_extra = True
                pool._add_request(self)
                return False
            elif pool._getmode == POOL_GETMODE_NOWAIT:
                errors._raise_err(errors.ERR_POOL_NO_CONNECTION_AVAILABLE)

        # the pool has room to grow; ask the background task to create a
        # specific connection if the connection class does not match;
        # otherwise, ask the pool to grow (if it is not already growing)
        elif self.cclass_matches:
            if pool._num_to_create == 0:
                pool._num_to_create = min(pool.increment,
                                          pool.max - pool._open_count)

        # wait for the pool to grow or a connection to be returned to the pool
        pool._add_request(self)
        return False

    cdef int reject(self) except -1:
        """
        Called when a request has been rejected for any reason (such as when a
        wait timeout has been exceeded). Any connection that is associated with
        the request is returned to the pool or destroyed, depending on the
        request.
        """
        cdef:
            BaseThinPoolImpl pool_impl = self.pool_impl
            BaseThinConnImpl conn_impl = self.conn_impl
        if conn_impl is not None:
            self.conn_impl = None
            if conn_impl._is_pool_extra:
                conn_impl._is_pool_extra = False
                pool_impl._conn_impls_to_drop.append(conn_impl)
            elif conn_impl.invoke_session_callback:
                pool_impl._free_new_conn_impls.append(conn_impl)
            else:
                pool_impl._free_used_conn_impls.append(conn_impl)


cdef class PoolCloser:
    cdef:
        bint closing
        object lock
        set pools

    def __init__(self):
        self.lock = threading.Lock()
        self.closing = False

    cdef int add_pool(self, ThinPoolImpl pool_impl) except -1:
        """
        Adds a pool to the list of pools to close. If this is the first pool
        being added, a thread is started to ensure that all remaining open
        pools are closed gracefully at shutdown of the interpreter.
        """
        with self.lock:
            if self.pools is None:
                self.pools = set()
                threading.Thread(target=self.close_pools_gracefully).start()
            self.pools.add(pool_impl)

    def close_pools_gracefully(self):
        """
        Closes all remaining open pools gracefully. Since pools start a
        background thread, the interpreter will not shut down until these
        background threads have completed, nor will registered atexit functions
        run. Marking the background threads as daemon threads allows the
        interpreter to kill them but then pools are not closed gracefully. This
        method of starting a thread to wait until the main thread finishes
        handles all of these situations.
        """
        cdef ThinPoolImpl pool_impl
        threading.main_thread().join()      # wait for main thread to finish
        self.closing = True
        for pool_impl in self.pools:
            pool_impl._shutdown()

    cdef int remove_pool(self, ThinPoolImpl pool_impl) except -1:
        """
        Removes a pool from the list of pools to close.
        """
        with self.lock:
            if not self.closing:
                self.pools.remove(pool_impl)


cdef PoolCloser pool_closer = PoolCloser()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# oci_config_provider.py
#
# Python file contains the hook method config_oci_hook() that fetches config
# store from OCI Object Storage.
# -----------------------------------------------------------------------------

import base64
import json
import oci
import oracledb
import re

from urllib.parse import urlparse, parse_qs

oci_from_file = oci.config.from_file
oci_client_error = oci.exceptions.ClientError
oci_object_storage_client = oci.object_storage.ObjectStorageClient
oci_secrets_client = oci.secrets.SecretsClient


"""
Pattern to parse OCI Object Connect String
"""
cloud_net_naming_pattern_oci = re.compile(
    r"(?P<objservername>[^/]+)/n/(?P<namespace>[^/]+)/b/(?P<bucketname>[^/]+)/o/(?P<filename>[^/]+)(/c/(?P<alias>[^/]+))?"
)


def _get_config(parameters, connect_params):
    config = {}

    credential, signer = _get_credential(parameters)
    auth_method = parameters.get("authentication")
    if auth_method is not None:
        auth_method = auth_method.upper()

    if auth_method is None or auth_method == "OCI_DEFAULT":
        client_oci = oci_object_storage_client(credential)
    elif (
        auth_method == "OCI_INSTANCE_PRINCIPAL"
        or auth_method == "OCI_RESOURCE_PRINCIPAL"
    ):
        client_oci = oci_object_storage_client(
            config=credential, signer=signer
        )
    get_object_request = {
        "object_name": _get_required_parameter(parameters, "filename"),
        "bucket_name": _get_required_parameter(parameters, "bucketname"),
        "namespace_name": _get_required_parameter(parameters, "namespace"),
    }

    get_object_response = client_oci.get_object(**get_object_request)
    resp = _stream_to_string(get_object_response.data)
    settings = json.loads(resp)
    user_alias = parameters.get("alias")
    if user_alias:
        settings = settings[user_alias]

    # Connect Descriptor
    config["connect_descriptor"] = _get_required_parameter(
        settings, "connect_descriptor"
    )

    # user and password
    if connect_params.user is None:
        config["user"] = settings.get("user")
        if "password" in settings:
            config["password"] = pwd = settings["password"]
            if pwd["type"] == "ocivault":
                authentication = pwd.setdefault("authentication", {})
                authentication.setdefault("method", auth_method)
                authentication["credential"] = credential

    # config cache settings
    config["config_time_to_live"] = settings.get("config_time_to_live")
    config["config_time_to_live_grace_period"] = settings.get(
        "config_time_to_live_grace_period"
    )

    # pyo parameters settings
    config["pyo"] = settings.get("pyo", None)

    # set the configuration
    connect_params.set_from_config(config)


def _get_credential(parameters):
    """
    Returns the appropriate credential given the input supplied by the original
    connect string.
    """
    auth_method = parameters.get("authentication", parameters.get("method"))

    if auth_method is not None:
        auth_method = auth_method.upper()

    # if region is not in connection string, retrieve from object server name.
    region = parameters.get(
        "oci_region", _retrieve_region(parameters.get("objservername"))
    )

    try:
        if auth_method is None or auth_method == "OCI_DEFAULT":
            # Default Authentication
            # default path ~/.oci/config
            return oci_from_file(), None
    except oci.exceptions.ClientError:
        # try to create config with connection string parameters.
        if "oci_tenancy" in parameters and "oci_user" in parameters:
            with open(parameters["oci_key_file"], "r") as file_content:
                public_key = file_content.read()
            provider = dict(
                tenancy=parameters["oci_tenancy"],
                user=parameters["oci_user"],
                fingerprint=parameters["oci_fingerprint"],
                key_file=parameters["oci_key_file"],
                private_key_content=public_key,
                region=region,
            )
            return provider, None

    if auth_method == "OCI_INSTANCE_PRINCIPAL":
        signer = oci.auth.signers.InstancePrincipalsSecurityTokenSigner()
        return (
            dict(region=region),
            signer,
        )

    elif auth_method == "OCI_RESOURCE_PRINCIPAL":
        signer = oci.auth.signers.get_resource_principals_signer()
        return {}, signer
    else:
        msg = "Authentication options were not available in Connection String"
        raise Exception(msg)


def _get_required_parameter(parameters, name, location="connection string"):
    try:
        return parameters[name]
    except KeyError:
        message = f'Parameter named "{name}" is missing from {location}'
        raise Exception(message) from None


def _parse_parameters(protocol_arg: str) -> dict:
    """
    Parse the parameters from the protocol argument string.
    """
    pos = protocol_arg.find("?")
    parsed_url = urlparse(protocol_arg[pos + 1 :])
    parsed_values = parse_qs(parsed_url.path)
    parameters = {
        key.lower(): value[0] for key, value in parsed_values.items()
    }

    match = cloud_net_naming_pattern_oci.match(protocol_arg[:pos])
    if match:
        parameters["objservername"] = match.group("objservername")
        parameters["namespace"] = match.group("namespace")
        parameters["bucketname"] = match.group("bucketname")
        parameters["filename"] = match.group("filename")
        if match.group("alias"):
            parameters["alias"] = match.group("alias")
    return parameters


def password_type_oci_vault_hook(args):
    secret_id = _get_required_parameter(
        args, "value", '"password" key section'
    )
    authentication = args.get("authentication")
    if authentication is None:
        raise Exception(
            "OCI Key Vault authentication details were not provided."
        )

    # if credentials are not present, create credentials with given
    # authentication details.
    credential = authentication.get("credential")
    if credential is None:
        credential, signer = _get_credential(authentication)

    auth_method = authentication.get("method")
    if auth_method is not None:
        auth_method = auth_method.upper()
    if auth_method is None or auth_method == "OCI_DEFAULT":
        secret_client_oci = oci_secrets_client(credential)
    elif auth_method == "OCI_INSTANCE_PRINCIPAL":
        signer = oci.auth.signers.InstancePrincipalsSecurityTokenSigner()
        secret_client_oci = oci_secrets_client(
            config=credential, signer=signer
        )
    elif auth_method == "OCI_RESOURCE_PRINCIPAL":
        signer = oci.auth.signers.get_resource_principals_signer()
        secret_client_oci = oci_secrets_client(
            config=credential, signer=signer
        )

    get_secret_bundle_response = secret_client_oci.get_secret_bundle(
        secret_id=secret_id
    )
    # decoding the vault content
    b64content = get_secret_bundle_response.data.secret_bundle_content.content
    return base64.b64decode(b64content).decode()


def _retrieve_region(objservername):
    if objservername is not None:
        arr = objservername.split(".")
        return arr[1].lower().replace("_", "-")


def _stream_to_string(stream):
    return b"".join(stream).decode()


def config_oci_hook(
    protocol: str, protocol_arg: str, connect_params: oracledb.ConnectParams
):
    """
    Hook for handling parameters stored in an OCI Object store.
    """
    parameters = _parse_parameters(protocol_arg)
    _get_config(parameters, connect_params)


oracledb.register_password_type("ocivault", password_type_oci_vault_hook)
oracledb.register_protocol("config-ociobject", config_oci_hook)



# -----------------------------------------------------------------------------
# Copyright (c) 2022, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
4600 - Module for testing the handling of type changes in queries.
"""

import datetime
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    def __test_type_change(
        self,
        query_frag_1,
        query_value_1,
        query_frag_2,
        query_value_2,
        table_name="dual",
        type_handler=None,
    ):
        if test_env.get_is_implicit_pooling():
            self.skipTest("sessions can change with implicit pooling")
        orig_type_handler = self.conn.outputtypehandler
        self.conn.outputtypehandler = type_handler
        try:
            self.cursor.execute(
                f"""
                create or replace view TestTypesChanged as
                select {query_frag_1} as value
                from {table_name}
                """
            )
            self.cursor.execute("select * from TestTypesChanged")
            self.assertEqual(self.cursor.fetchall(), [(query_value_1,)])
            self.cursor.execute(
                f"""
                create or replace view TestTypesChanged as
                    select {query_frag_2} as value
                    from dual
                """
            )
            self.cursor.execute("select * from TestTypesChanged")
            self.assertEqual(self.cursor.fetchall(), [(query_value_2,)])
        finally:
            self.conn.outputtypehandler = orig_type_handler

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4600(self):
        "4600 - test data type changing from VARCHAR to CLOB"
        self.__test_type_change(
            "cast('string_4600' as VARCHAR2(15))",
            "string_4600",
            "to_clob('clob_4600')",
            "clob_4600",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4601(self):
        "4601 - test data type changing from CHAR to CLOB"
        self.__test_type_change(
            "cast('string_4601' as CHAR(11))",
            "string_4601",
            "to_clob('clob_4601')",
            "clob_4601",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4602(self):
        "4602 - test data type changing from LONG to CLOB"
        self.cursor.execute("truncate table TestLongs")
        self.cursor.execute("insert into TestLongs values (1, 'string_4602')")
        self.__test_type_change(
            "LongCol",
            "string_4602",
            "to_clob('clob_4602')",
            "clob_4602",
            "TestLongs",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4603(self):
        "4603 - test data type changing from NVARCHAR to CLOB"
        self.__test_type_change(
            "cast('string_4603' as NVARCHAR2(15))",
            "string_4603",
            "to_clob('clob_4603')",
            "clob_4603",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4604(self):
        "4604 - test data type changing from NCHAR to CLOB"
        self.__test_type_change(
            "cast('string_4604' as NCHAR(11))",
            "string_4604",
            "to_clob('clob_4604')",
            "clob_4604",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4605(self):
        "4605 - test data type changing from RAW to BLOB"
        self.__test_type_change(
            "utl_raw.cast_to_raw('string_4605')",
            b"string_4605",
            "to_blob(utl_raw.cast_to_raw('blob_4605'))",
            b"blob_4605",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4606(self):
        "4606 - test data type changing from LONGRAW to BLOB"
        self.cursor.execute("truncate table TestLongRaws")
        data = [1, b"string_4606"]
        self.cursor.execute("insert into TestLongRaws values (:1, :2)", data)
        self.__test_type_change(
            "LongRawCol",
            b"string_4606",
            "to_blob(utl_raw.cast_to_raw('blob_4606'))",
            b"blob_4606",
            "TestLongRaws",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4607(self):
        "4607 - test data type changing from VARCHAR to NCLOB"
        self.__test_type_change(
            "cast('string_4607' as VARCHAR2(15))",
            "string_4607",
            "to_nclob('nclob_4607')",
            "nclob_4607",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4608(self):
        "4608 - test data type changing from CHAR to NCLOB"
        self.__test_type_change(
            "cast('string_4608' as CHAR(11))",
            "string_4608",
            "to_nclob('nclob_4608')",
            "nclob_4608",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4609(self):
        "4609 - test data type changing from LONG to NCLOB"
        self.cursor.execute("truncate table TestLongs")
        self.cursor.execute("insert into TestLongs values (1, 'string_4609')")
        self.__test_type_change(
            "LongCol",
            "string_4609",
            "to_nclob('nclob_4609')",
            "nclob_4609",
            "TestLongs",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4610(self):
        "4610 - test data type changing from NVARCHAR to NCLOB"
        self.__test_type_change(
            "cast('string_4610' as NVARCHAR2(15))",
            "string_4610",
            "to_nclob('nclob_4610')",
            "nclob_4610",
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4611(self):
        "4611 - test data type changing from NCHAR to NCLOB"
        self.__test_type_change(
            "cast('string_4611' as NCHAR(11))",
            "string_4611",
            "to_nclob('nclob_4611')",
            "nclob_4611",
        )

    def test_4612(self):
        "4612 - test data type changing from VARCHAR to NUMBER"
        self.__test_type_change(
            "cast('string_4612' as VARCHAR2(15))",
            "string_4612",
            "to_number('4612')",
            4612,
        )

    def test_4613(self):
        "4613 - test data type changing from NUMBER to VARCHAR"
        self.__test_type_change(
            "to_number('4613')",
            4613,
            "cast('string_4613' as VARCHAR2(15))",
            "string_4613",
        )

    def test_4614(self):
        "4614 - test data type changing from STRING to DATE"
        self.__test_type_change(
            "cast('string_4614' as VARCHAR2(15))",
            "string_4614",
            "to_date('04-JAN-2022')",
            datetime.datetime(2022, 1, 4, 0, 0),
        )

    def test_4615(self):
        "4615 - test data type changing from DATE to STRING"
        self.__test_type_change(
            "to_date('04-JAN-2022')",
            datetime.datetime(2022, 1, 4, 0, 0),
            "cast('string_4615' as VARCHAR2(15))",
            "string_4615",
        )

    def test_4616(self):
        "4616 - test data type changing from NUMBER to DATE"
        self.__test_type_change(
            "to_number('4616')",
            4616,
            "to_date('05-JAN-2022')",
            datetime.datetime(2022, 1, 5, 0, 0),
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4617(self):
        "4617 - test data type changing from CLOB to VARCHAR"

        def type_handler(cursor, metadata):
            if metadata.type_code is oracledb.DB_TYPE_CLOB:
                return cursor.var(
                    oracledb.DB_TYPE_VARCHAR,
                    size=32768,
                    arraysize=cursor.arraysize,
                )

        self.__test_type_change(
            "to_clob('clob_4617')",
            "clob_4617",
            "cast('string_4617' as VARCHAR2(15))",
            "string_4617",
            type_handler=type_handler,
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4618(self):
        "4618 - test data type changing from NCLOB to NVARCHAR"

        def type_handler(cursor, metadata):
            if metadata.type_code is oracledb.DB_TYPE_NCLOB:
                return cursor.var(
                    oracledb.DB_TYPE_NVARCHAR,
                    size=32768,
                    arraysize=cursor.arraysize,
                )

        self.__test_type_change(
            "to_nclob('nclob_4618')",
            "nclob_4618",
            "cast('nstring_4618' as NVARCHAR2(15))",
            "nstring_4618",
            type_handler=type_handler,
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4619(self):
        "4619 - test data type changing from CLOB to NVARCHAR"

        def type_handler(cursor, metadata):
            if metadata.type_code is oracledb.DB_TYPE_CLOB:
                return cursor.var(
                    oracledb.DB_TYPE_NVARCHAR,
                    size=32768,
                    arraysize=cursor.arraysize,
                )

        self.__test_type_change(
            "to_clob('clob_4619')",
            "clob_4619",
            "cast('string_4619' as VARCHAR2(15))",
            "string_4619",
            type_handler=type_handler,
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4620(self):
        "4620 - test data type changing from BLOB to RAW"

        def type_handler(cursor, metadata):
            if metadata.type_code is oracledb.DB_TYPE_BLOB:
                return cursor.var(
                    oracledb.DB_TYPE_RAW,
                    size=32768,
                    arraysize=cursor.arraysize,
                )

        self.__test_type_change(
            "to_blob(utl_raw.cast_to_raw('blob_4620'))",
            b"blob_4620",
            "utl_raw.cast_to_raw('string_4620')",
            b"string_4620",
            type_handler=type_handler,
        )

    @unittest.skipIf(
        not test_env.get_is_thin(),
        "thick mode doesn't support this type change",
    )
    def test_4621(self):
        "4621 - test data type changing from NVARCHAR to CLOB"
        self.__test_type_change(
            "cast('string_4621' as NVARCHAR2(15))",
            "string_4621",
            "to_clob('clob_4621')",
            "clob_4621",
        )


if __name__ == "__main__":
    test_env.run_test_cases()
